@@ -1,31 +1,87 @@
+import logging
+import itertools
+
 import zigpy_znp.types as t
 import zigpy_znp.commands as c
 from zigpy_znp.types import nvids
 from zigpy_znp.exceptions import SecurityError, InvalidCommandResponse
 
+LOGGER = logging.getLogger(__name__)
+
+
 # Some NVIDs don't really exist and Z-Stack doesn't behave consistently when operations
 # are performed on them.
 PROXIED_NVIDS = {nvids.OsalNvIds.POLL_RATE_OLD16}
 
 
-def serialize(value) -> bytes:
-    if hasattr(value, "serialize"):
-        value = value.serialize()
-    elif not isinstance(value, (bytes, bytearray)):
-        raise TypeError(
-            f"Only bytes or serializable types can be written to NVRAM."
-            f" Got {value!r} (type {type(value)})"
-        )
-
-    if not value:
-        raise ValueError("NVRAM value cannot be empty")
-
-    return value
-
-
 class NVRAMHelper:
     def __init__(self, znp):
         self.znp = znp
+        self.align_structs = None
+
+    async def determine_alignment(self) -> None:
+        """
+        Automatically determine struct memory alignment. Must be called before any
+        structs are read/written.
+        """
+
+        # TODO: Figure out a way to use `c.UTIL.AssocFindDevice.Req(Index=0)`!
+        #       This is the only (known) MT command to just send a struct's in-memory
+        #       representation over serial.
+
+        # NWKKEY is almost always present, regardless of adapter state.
+        # It is an 8-bit sequence number, a 16 byte key, and a 32-bit frame counter.
+        # This is at least 1 + 16 + 4 = 21 bytes, or 24 if you have to 32-bit align.
+        value = await self.osal_read(nvids.OsalNvIds.NWKKEY, item_type=t.Bytes)
+
+        if len(value) == 24:
+            self.align_structs = True
+        elif len(value) == 21:
+            self.align_structs = False
+        else:
+            raise ValueError(f"Unexpected value for NWKKEY: {value!r}")
+
+        LOGGER.debug("Detected struct alignment: %s", self.align_structs)
+
+    def serialize(self, value) -> bytes:
+        """
+        Serialize an object, automatically computing struct padding based on the target
+        platform.
+        """
+
+        if hasattr(value, "serialize"):
+            if isinstance(value, (t.CStruct, t.BaseListType)):
+                assert self.align_structs is not None
+                value = value.serialize(align=self.align_structs)
+            else:
+                value = value.serialize()
+        elif not isinstance(value, (bytes, bytearray)):
+            raise TypeError(
+                f"Only bytes or serializable types can be written to NVRAM."
+                f" Got {value!r} (type {type(value)})"
+            )
+
+        if not value:
+            raise ValueError("NVRAM value cannot be empty")
+
+        return value
+
+    def deserialize(self, data: bytes, item_type, *, allow_trailing=False):
+        """
+        Serialize raw bytes, automatically computing struct padding based on the target
+        platform.
+        """
+
+        if issubclass(item_type, (t.CStruct, t.BaseListType)):
+            assert self.align_structs is not None
+            value, remaining = item_type.deserialize(data, align=self.align_structs)
+        else:
+            value, remaining = item_type.deserialize(data)
+
+        if remaining and not allow_trailing:
+            raise ValueError(f"Data left after deserialization: {remaining!r}")
+
+        return value
 
     async def osal_delete(self, nv_id: t.uint16_t) -> bool:
         """
@@ -51,7 +107,7 @@ class NVRAMHelper:
         Serializes all serializable values and passes bytes directly.
         """
 
-        value = serialize(value)
+        value = self.serialize(value)
         length = (await self.znp.request(c.SYS.OSALNVLength.Req(Id=nv_id))).ItemLen
 
         # Recreate the item if the length is not correct
@@ -91,7 +147,7 @@ class NVRAMHelper:
                 RspStatus=t.Status.SUCCESS,
             )
 
-    async def osal_read(self, nv_id: t.uint16_t) -> bytes:
+    async def osal_read(self, nv_id: t.uint16_t, *, item_type):
         """
         Reads a complete value from NVRAM.
 
@@ -105,7 +161,10 @@ class NVRAMHelper:
                 RspStatus=t.Status.SUCCESS,
             )
 
-            return read_rsp.Value
+            value = self.deserialize(read_rsp.Value, item_type)
+            LOGGER.debug('Read NVRAM["LEGACY"][0x%04x] = %r', nv_id, value)
+
+            return value
 
         # Every item has a length, even missing ones
         length = (await self.znp.request(c.SYS.OSALNVLength.Req(Id=nv_id))).ItemLen
@@ -130,7 +189,7 @@ class NVRAMHelper:
             # Not all items can be read out due to security policies, though this can
             # easily be bypassed for some. The SAPI "ConfigId" is only 8 bits which
             # means some nvids are not able to read this way.
-            if not self.znp.capabilities & t.MTCapabilities.CAP_SAPI or nv_id > 0xFF:
+            if not self.znp.capabilities & t.MTCapabilities.SAPI or nv_id > 0xFF:
                 raise SecurityError(
                     f"NV item cannot be read due to security constraints: {nv_id!r}"
                 )
@@ -145,10 +204,17 @@ class NVRAMHelper:
 
         assert len(data) == length
 
-        return data
+        value = self.deserialize(data, item_type)
+        LOGGER.debug('Read NVRAM["LEGACY"][0x%04x] = %r', nv_id, value)
+
+        return value
 
     async def delete(
-        self, sys_id: t.uint8_t, item_id: t.uint16_t, sub_id: t.uint16_t
+        self,
+        *,
+        sys_id: t.uint8_t = nvids.NvSysIds.ZSTACK,
+        item_id: t.uint16_t,
+        sub_id: t.uint16_t,
     ) -> bool:
         """
         Deletes a subitem from NVRAM. Returns whether or not the item existed.
@@ -162,13 +228,13 @@ class NVRAMHelper:
 
     async def write(
         self,
-        sys_id: t.uint8_t,
+        *,
+        sys_id: t.uint8_t = nvids.NvSysIds.ZSTACK,
         item_id: t.uint16_t,
         sub_id: t.uint16_t,
         value,
-        *,
         create: bool = True,
-    ):
+    ) -> None:
         """
         Writes a value to NVRAM for the specified subsystem, item, and subitem.
 
@@ -176,7 +242,7 @@ class NVRAMHelper:
         NVWrite(sys_id=ZSTACK, item_id=LEGACY, sub_id=1) in the background.
         """
 
-        value = serialize(value)
+        value = self.serialize(value)
         length = (
             await self.znp.request(
                 c.SYS.NVLength.Req(SysId=sys_id, ItemId=item_id, SubId=sub_id)
@@ -232,7 +298,12 @@ class NVRAMHelper:
             )
 
     async def read(
-        self, sys_id: t.uint8_t, item_id: t.uint16_t, sub_id: t.uint16_t
+        self,
+        *,
+        sys_id: t.uint8_t = nvids.NvSysIds.ZSTACK,
+        item_id: t.uint16_t,
+        sub_id: t.uint16_t,
+        item_type,
     ) -> bytes:
         """
         Reads a value from NVRAM for the specified subsystem, item, and subitem.
@@ -272,4 +343,80 @@ class NVRAMHelper:
 
         assert len(data) == length
 
-        return data
+        value = self.deserialize(data, item_type)
+        LOGGER.debug("Read NVRAM[%s][%s][0x%04x] = %r", sys_id, item_id, sub_id, value)
+
+        return value
+
+    async def write_table(
+        self,
+        *,
+        sys_id: t.uint8_t = nvids.NvSysIds.ZSTACK,
+        item_id: t.uint16_t,
+        values,
+        fill_value,
+    ) -> None:
+        for sub_id, value in itertools.zip_longest(
+            range(0x0000, 0xFFFF + 1), values, fillvalue=fill_value
+        ):
+            value = self.serialize(value)
+
+            try:
+                await self.write(
+                    sys_id=sys_id,
+                    item_id=item_id,
+                    sub_id=sub_id,
+                    value=value,
+                    create=False,
+                )
+            except KeyError:
+                break
+
+    async def osal_write_table(
+        self, start_nvid: t.uint16_t, end_nvid: t.uint16_t, values, *, fill_value
+    ) -> None:
+        values = list(values)
+
+        for nvid, value in itertools.zip_longest(
+            range(start_nvid, end_nvid + 1), values, fillvalue=fill_value
+        ):
+            value = self.serialize(value)
+
+            try:
+                await self.osal_write(
+                    nv_id=nvid,
+                    value=value,
+                    create=False,
+                )
+            except KeyError:
+                break
+
+    async def read_table(
+        self,
+        *,
+        sys_id: t.uint8_t = nvids.NvSysIds.ZSTACK,
+        item_id: t.uint16_t,
+        item_type=t.Bytes,
+    ):
+        for sub_id in range(0x0000, 0xFFFF + 1):
+            try:
+                yield await self.read(
+                    sys_id=sys_id,
+                    item_id=item_id,
+                    sub_id=sub_id,
+                    item_type=item_type,
+                )
+            except KeyError:
+                break
+
+    async def osal_read_table(
+        self,
+        start_nvid: t.uint16_t,
+        end_nvid: t.uint16_t,
+        item_type=t.Bytes,
+    ):
+        for nvid in range(start_nvid, end_nvid + 1):
+            try:
+                yield await self.osal_read(nvid, item_type=item_type)
+            except KeyError:
+                break
